"""
Authorization scopes for arXiv users and clients.

The concept of authorization scope comes from OAuth 2.0 (`RFC 6749 §3.3
<https://tools.ietf.org/html/rfc6749#section-3.3>`_). For a nice primer, see
`this blog post <https://brandur.org/oauth-scope>`_. The basic idea is that
the authorization associated with an access token can be limited, e.g. to
limit what actions an API client can take on behalf of a user.

In this package, the scope concept is applied to both API client and end-user
sessions. When the session is created, we consult the relevant bits of data in
our system (e.g. what roles the user has, what privileges are associated with
those roles) to determine what the user is authorized to do. Those privileges
are attached to the user's session as authorization scopes.

This module simply defines a set of constants (str) that can be used to refer
to specific authorization scopes. Rather than refer to scopes by writing new
str objects, these constants should be imported and used. This improves the
semantics of code, and reduces the risk of programming errors. For an example,
see :mod:`arxiv.users.auth.decorators`.

"""
from typing import Optional
from ..domain import Scope

READ_PUBLIC = Scope(Scope.domains.PUBLIC, Scope.actions.READ)
"""
Authorizes access to public endpoints.

This is implicitly granted to all anonymous users. For endpoints requiring
authentication (e.g. APIs) this scope can be used to denote baseline read
access for clients.
"""

EDIT_PROFILE = Scope(Scope.domains.PROFILE, Scope.actions.UPDATE)
"""
Authorizes editing user profile.

This includes things like affiliation, full name, etc..
"""

VIEW_PROFILE = Scope(Scope.domains.PROFILE, Scope.actions.READ)
"""
Authorizes viewing the content of a user profile.

This includes things like affiliation, full name, and e-mail address.
"""

CREATE_SUBMISSION = Scope(Scope.domains.SUBMISSION, Scope.actions.CREATE)
"""Authorizes creating a new submission."""

EDIT_SUBMISSION = Scope(Scope.domains.SUBMISSION, Scope.actions.UPDATE)
"""Authorizes updating a submission that has not yet been announced."""

VIEW_SUBMISSION = Scope(Scope.domains.SUBMISSION, Scope.actions.READ)
"""Authorizes viewing a submission."""

PROXY_SUBMISSION = Scope(Scope.domains.SUBMISSION, Scope.actions.PROXY)
"""
Authorizes creating a submission on behalf of another user.

This authorization is specifically for human users submitting on behalf of
other human users. For client authorization to submit on behalf of a user,
<code>submission:create</code> should be used instead.
"""

READ_UPLOAD = Scope(Scope.domains.UPLOAD, Scope.actions.READ)
"""Authorizes viewing the content of an upload workspace."""

WRITE_UPLOAD = Scope(Scope.domains.UPLOAD, Scope.actions.UPDATE)
"""Authorizes uploading files to to a workspace."""

RELEASE_UPLOAD = Scope(Scope.domains.UPLOAD, Scope.actions.RELEASE)
"""Authorizes releasing an upload workspace."""

DELETE_UPLOAD_WORKSPACE = Scope(Scope.domains.UPLOAD, 'delete_workspace')
DELETE_UPLOAD_FILE = Scope(Scope.domains.UPLOAD, Scope.actions.DELETE)

READ_UPLOAD_LOGS = Scope(Scope.domains.UPLOAD, 'read_logs')
READ_UPLOAD_SERVICE_LOGS = Scope(Scope.domains.UPLOAD, 'read_service_logs')

GENERAL_USER = [
    EDIT_PROFILE,
    VIEW_PROFILE,
    CREATE_SUBMISSION,
    EDIT_SUBMISSION,
    VIEW_SUBMISSION
]

_HUMAN_LABELS = {
    EDIT_PROFILE: "Grants authorization to change the contents of your user"
                  " profile. This includes your affiliation, preferred name,"
                  " default submission category, etc.",
    VIEW_PROFILE: "Grants authorization to view the contents of your user"
                  " profile. This includes your affiliation, preferred name,"
                  " default submission category, etc.",
    CREATE_SUBMISSION: "Grants authorization to submit papers on your behalf.",
    EDIT_SUBMISSION: "Grants authorization to make changes to your submissions"
                     " that have not yet been announced. For example, to"
                     " update the DOI or journal reference field on your"
                     " behalf. Note that this affects only the metadata of"
                     " your submission, and not the content.",
    VIEW_SUBMISSION: "Grants authorization to view your submissions, including"
                     " those that have not yet been announced. Note that this"
                     " only applies to the submission metadata, and not to the"
                     " uploaded submission content.",
    READ_UPLOAD: "Grants authorization to view the contents of your uploads.",
    WRITE_UPLOAD: "Grants authorization to add and delete files on your"
                  " behalf.",
}

def get_human_label(scope: str) -> Optional[str]:
    """The the human-readable label for a scope, for display to end users."""
    return _HUMAN_LABELS.get(scope)



"""Helpers and utilities for :mod:`arxiv.users`."""

import os
from typing import List
from pytz import timezone
import uuid
from datetime import timedelta, datetime
from arxiv.users import auth, domain


def generate_token(user_id: str, email: str, username: str,
                   first_name: str = 'Jane', last_name: str = 'Doe',
                   suffix_name: str = 'IV',
                   affiliation: str = 'Cornell University',
                   rank: int = 3,
                   country: str = 'us',
                   default_category: domain.Category = (
                       domain.Category('astro-ph', 'GA')
                   ),
                   submission_groups: str = 'grp_physics',
                   endorsements: List[domain.Category] = [],
                   scope: List[domain.Scope] = []) -> None:
    """Generate an auth token for dev/testing purposes."""
    # Specify the validity period for the session.
    start = datetime.now(tz=timezone('US/Eastern'))
    end = start + timedelta(seconds=36000)   # Make this as long as you want.

    # Create a user with endorsements in astro-ph.CO and .GA.
    session = domain.Session(
        session_id=str(uuid.uuid4()),
        start_time=start, end_time=end,
        user=domain.User(
            user_id=user_id,
            email=email,
            username=username,
            name=domain.UserFullName(first_name, last_name, suffix_name),
            profile=domain.UserProfile(
                affiliation=affiliation,
                rank=int(rank),
                country=country,
                default_category=default_category,
                submission_groups=submission_groups.split(',')
            )
        ),
        authorizations=domain.Authorizations(scopes=scope,
                                             endorsements=endorsements)
    )
    token = auth.tokens.encode(session, os.environ['JWT_SECRET'])
    return token



"""
Internal service API for the distributed session store.

Used to create, delete, and verify user and client session.
"""

import json
import time
import uuid
import random
from datetime import datetime, timedelta
import dateutil.parser
from pytz import timezone

from typing import Any, Optional, Union, Tuple

from functools import wraps
import redis
import rediscluster

import jwt

from ... import domain
from ..exceptions import SessionCreationFailed, InvalidToken, \
    SessionDeletionFailed, UnknownSession, ExpiredToken

from arxiv.base.globals import get_application_config, get_application_global
from arxiv.base import logging

logger = logging.getLogger(__name__)
EASTERN = timezone('US/Eastern')


def _generate_nonce(length: int = 8) -> str:
    return ''.join([str(random.randint(0, 9)) for i in range(length)])


class SessionStore(object):
    """
    Manages a connection to Redis.

    In fact, the StrictRedis instance is thread safe and connections are
    attached at the time a command is executed. This class simply provides a
    container for configuration.
    """

    def __init__(self, host: str, port: int, db: int, secret: str,
                 duration: int = 7200, token: str = None,
                 cluster: bool = True) -> None:
        """Open the connection to Redis."""
        # params = #, db=db)
        logger.debug('New Redis connection at %s, port %s', host, port)
        if cluster:
            self.r = rediscluster.StrictRedisCluster(
                startup_nodes=[{'host': host, 'port': str(port)}],
                skip_full_coverage_check=True
            )
        else:
            self.r = redis.StrictRedis(host=host, port=port)
        self._secret = secret
        self._duration = duration

    def create(self, authorizations: domain.Authorizations,
               ip_address: str, remote_host: str, tracking_cookie: str = '',
               user: Optional[domain.User] = None,
               client: Optional[domain.Client] = None,
               session_id: Optional[str] = None) -> domain.Session:
        """
        Create a new session.

        Parameters
        ----------
        authorizations : :class:`domain.Authorizations`
        ip_address : str
        remote_host : str
        tracking_cookie : str
        user : :class:`domain.User`
        client : :class:`domain.Client`

        Returns
        -------
        :class:`.Session`
        """
        if session_id is None:
            session_id = str(uuid.uuid4())
        start_time = datetime.now(tz=EASTERN)
        end_time = start_time + timedelta(seconds=self._duration)
        session = domain.Session(
            session_id=session_id,
            user=user,
            client=client,
            start_time=start_time,
            end_time=end_time,
            authorizations=authorizations,
            nonce=_generate_nonce()
        )

        try:
            self.r.set(session_id, self._encode(domain.to_dict(session)),
                       ex=self._duration)
        except redis.exceptions.ConnectionError as e:
            raise SessionCreationFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionCreationFailed(f'Failed to create: {e}') from e

        return session

    def generate_cookie(self, session: domain.Session) -> str:
        """Generate a cookie from a :class:`domain.Session`."""
        return self._pack_cookie({
            'user_id': session.user.user_id,
            'session_id': session.session_id,
            'nonce': session.nonce,
            'expires': session.end_time.isoformat()
        })

    def delete(self, cookie: str) -> None:
        """
        Delete a session.

        Parameters
        ----------
        cookie : str
        """
        cookie_data = self._unpack_cookie(cookie)
        self.delete_by_id(cookie_data['session_id'])

    def delete_by_id(self, session_id: str) -> None:
        """
        Delete a session in the key-value store by ID.

        Parameters
        ----------
        session_id : str
        """
        try:
            self.r.delete(session_id)
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e

    def validate_session_against_cookie(self, session: domain.Session,
                                        cookie: str) -> None:
        """
        Validate session data against a cookie.

        Parameters
        ----------
        session : :class:`Session`
        cookie : str

        Raises
        ------
        :class:`InvalidToken`
            Raised if the data in the cookie does not match the session data.
        """
        cookie_data = self._unpack_cookie(cookie)
        if cookie_data['nonce'] != session.nonce \
                or session.user.user_id != cookie_data['user_id']:
            raise InvalidToken('Invalid token; likely a forgery')

    def load(self, cookie: str, decode: bool = True) -> domain.Session:
        """Load a session using a session cookie."""
        try:
            cookie_data = self._unpack_cookie(cookie)
            expires = dateutil.parser.parse(cookie_data['expires'])
        except (KeyError, jwt.exceptions.DecodeError) as e:    # type: ignore
            raise InvalidToken('Token payload malformed') from e

        if expires <= datetime.now(tz=EASTERN):
            raise InvalidToken('Session has expired')

        session = self.load_by_id(cookie_data['session_id'], decode=decode)
        if session.expired:
            raise ExpiredToken('Session has expired')
        if session.user is None and session.client is None:
            raise InvalidToken('Neither user nor client data are present')

        self.validate_session_against_cookie(session, cookie)
        return session

    def load_by_id(self, session_id: str, decode: bool = True) \
            -> Union[domain.Session, str]:
        """Get session data by session ID."""
        session_jwt: str = self.r.get(session_id)
        if not session_jwt:
            logger.error(f'No such session: {session_id}')
            raise UnknownSession(f'Failed to find session {session_id}')
        if decode:
            return self._decode(session_jwt)
        return session_jwt

    def _encode(self, session_data: dict) -> str:
        return jwt.encode(session_data, self._secret)

    def _decode(self, session_jwt: str) -> domain.Session:
        try:
            session: domain.Session = domain.from_dict(
                domain.Session,
                jwt.decode(session_jwt, self._secret)
            )
        except jwt.exceptions.InvalidSignatureError:
            raise InvalidToken('Invalid or corrupted session token')
        return session

    def _unpack_cookie(self, cookie: str) -> dict:
        secret = self._secret
        try:
            data = dict(jwt.decode(cookie, secret))
        except jwt.exceptions.DecodeError as e:   # type: ignore
            raise InvalidToken('Session cookie is malformed') from e
        return data

    def _pack_cookie(self, cookie_data: dict) -> str:
        secret = self._secret
        return jwt.encode(cookie_data, secret).decode('ascii')


def init_app(app: object = None) -> None:
    """Set default configuration parameters for an application instance."""
    config = get_application_config(app)
    config.setdefault('REDIS_HOST', 'localhost')
    config.setdefault('REDIS_PORT', '7000')
    config.setdefault('REDIS_DATABASE', '0')
    config.setdefault('REDIS_TOKEN', None)
    config.setdefault('REDIS_CLUSTER', '1')
    config.setdefault('JWT_SECRET', 'foosecret')
    config.setdefault('SESSION_DURATION', '7200')


def get_redis_session(app: object = None) -> SessionStore:
    """Get a new session with the search index."""
    config = get_application_config(app)
    host = config.get('REDIS_HOST', 'localhost')
    port = int(config.get('REDIS_PORT', '7000'))
    db = int(config.get('REDIS_DATABASE', '0'))
    token = config.get('REDIS_TOKEN', None)
    cluster = config.get('REDIS_CLUSTER', '1') == '1'
    secret = config['JWT_SECRET']
    duration = int(config.get('SESSION_DURATION', '7200'))
    return SessionStore(host, port, db, secret, duration, token=token,
                        cluster=cluster)


def current_session() -> SessionStore:
    """Get/create :class:`.SearchSession` for this context."""
    g = get_application_global()
    if not g:
        return get_redis_session()
    if 'redis' not in g:
        g.redis = get_redis_session()
    return g.redis      # type: ignore


@wraps(SessionStore.create)
def create(authorizations: domain.Authorizations,
           ip_address: str, remote_host: str, tracking_cookie: str = '',
           user: Optional[domain.User] = None,
           client: Optional[domain.Client] = None,
           session_id: Optional[str] = None) -> domain.Session:
    """Create a new session."""
    return current_session().create(authorizations, ip_address,
                                    remote_host, tracking_cookie,
                                    user=user, client=client,
                                    session_id=session_id)


@wraps(SessionStore.load)
def load(cookie: str, decode: bool = True) -> Union[domain.Session, str]:
    """Load a session by cookie value."""
    return current_session().load(cookie, decode=decode)


@wraps(SessionStore.load)
def load_by_id(session_id: str, decode: bool = True) \
        -> Union[domain.Session, str]:
    """Load a session by session ID."""
    return current_session().load_by_id(session_id, decode=decode)


@wraps(SessionStore.delete)
def delete(cookie: str) -> None:
    """Delete a session in the key-value store."""
    return current_session().delete(cookie)


@wraps(SessionStore.delete_by_id)
def delete_by_id(session_id: str) -> None:
    """Delete a session in the key-value store by ID."""
    return current_session().delete_by_id(session_id)


@wraps(SessionStore.generate_cookie)
def generate_cookie(session: domain.Session) -> str:
    """Generate a cookie from a :class:`domain.Session`."""
    return current_session().generate_cookie(session)



"""
Scope-based authorization of user/client requests.

This module provides :func:`scoped`, a decorator factory used to protect Flask
routes for which authorization is required. This is done by specifying a
required authorization scope (see :mod:`arxiv.users.auth.scopes`) and/or by
providing a custom authorizer function.

For routes that involve specific resources, a ``resource`` callback should also
be provided. That callback function should accept the same arguments as the
route function, and return the identifier for the resource as a string.

Using :func:`scoped` with an authorizer function allows you to define
application-specific authorization logic on a per-request basis without adding
complexity to request controllers. The call signature of the authorizer
function should be: ``(session: domain.Session, *args, **kwargs) -> bool``,
where `*args` and `**kwargs` are the positional and keyword arguments,
respectively, passed by Flask to the decorated route function (e.g. the
URL parameters).

.. note:: The authorizer function is only called if the session does not have
   a global or resource-specific instance of the required scope, or if a
   required scope is not specified.

Here's an example of how you might use this in a Flask application:

.. code-block:: python

   from arxiv.users.auth.decorators import scoped
   from arxiv.users.auth import scopes
   from arxiv.users import domain


   def is_owner(session: domain.Session, user_id: str, **kwargs) -> bool:
       '''Check whether the authenticated user matches the requested user.'''
       return session.user.user_id == user_id


   def get_resource_id(user_id: str) -> str:
       '''Get the user ID from the request.'''
       return user_id


   def redirect_to_login(user_id: str) -> Response:
       '''Send the unauthorized user to the log in page.'''
       return url_for('login')


   @blueprint.route('/<string:user_id>/profile', methods=['GET'])
   @scoped(scopes.EDIT_PROFILE, resource=get_resource_id,
           authorizer=user_is_owner, unauthorized=redirect_to_login)
   def edit_profile(user_id: str):
       '''User can update their account information.'''
       data, code, headers = profile.get_profile(user_id)
       return render_template('accounts/profile.html', **data)


When the decorated route function is called...

- If no session is available from either the middleware or the legacy database,
  the ``unauthorized`` callback is called, and/or :class:`Unauthorized`
  exception is raised.
- If a required scope was provided, the session is checked for the presence of
  that scope in this order:

  - Global scope (`:*`), e.g. for administrators.
  - Resource-specific scope (`:[resource_id]`), i.e. explicitly granted for a
    particular resource.
  - Generic scope (no resource part).

- If an authorization function was provided, the function is called only if
  a required scope was not provided, or if only the generic scope was found.
- Session data is added directly to the Flask request object as
  ``request.session``, for ease of access elsewhere in the application.
- Finally, if no exceptions have been raised, the route is called with the
  original parameters.

"""

from typing import Optional, Union, Callable, Any
from functools import wraps
from flask import request
from werkzeug.exceptions import BadRequest, Unauthorized, Forbidden
from arxiv import status
from arxiv.base import logging
from .. import domain

INVALID_TOKEN = {'reason': 'Invalid authorization token'}
INVALID_SCOPE = {'reason': 'Token not authorized for this action'}


logger = logging.getLogger(__name__)


def scoped(required: Optional[str] = None,
           resource: Optional[Callable] = None,
           authorizer: Optional[Callable] = None,
           unauthorized: Optional[Callable] = None) -> Callable:
    """
    Generate a decorator to enforce authorization requirements.

    Parameters
    ----------
    required : str
        The scope required on a user or client session in order use the
        decorated route. See :mod:`arxiv.users.auth.scopes`. If not provided,
        no scope will be enforced.
    resource : function
        If a route provides actions for a specific resource, a callable should
        be provided that accepts the route arguments and returns the resource
        identifier (str).
    authorizer : function
        In addition, an authorizer function may be passed to provide more
        specific authorization checks. For example, this function may check
        that the requesting user is the owner of a resource. Should have the
        signature: ``(session: domain.Session, *args, **kwargs) -> bool``.
        ``*args`` and ``**kwargs`` are the parameters passed to the decorated
        function. If the authorizer returns ``False``, an :class:`.`
        exception is raised.
    unauthorized : function
        A callback may be passed to handle cases in which the request is
        unauthorized. This function will be passed the same arguments as the
        original route function. If the callback returns anything other than
        ``None``, the return value will be treated as a response and execution
        will stop. Otherwise, an ``Unauthorized`` exception will be raised.
        If a callback is not provided (default) an ``Unauthorized`` exception
        will be raised.

    Returns
    -------
    function
        A decorator that enforces the required scope and calls the (optionally)
        provided authorizer.

    """
    def protector(func: Callable) -> Callable:
        """Decorator that provides scope enforcement."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Check the authorization token before executing the method.

            Will also raise exceptions passed by the auth middleware.

            Raises
            ------
            :class:`.Unauthorized`
                Raised when session data is not available.
            :class:`.`
                Raised when the session has insufficient auth scope, or the
                provided authorizer returns ``False``.

            """
            session = request.session
            scopes: List[domain.Scope] = []
            authorized: bool = False
            logger.debug('Required: %s, authorizer: %s, unauthorized: %s',
                         required, authorizer, unauthorized)
            # Use of the decorator implies that an auth session ought to be
            # present. So we'll complain here if it's not.
            if not session or not (session.user or session.client):
                logger.debug('No valid session; aborting')
                if unauthorized is not None:
                    response = unauthorized(*args, **kwargs)
                    if response is not None:
                        return response
                raise Unauthorized('Not a valid session')  # type: ignore

            if session.authorizations is not None:
                scopes = session.authorizations.scopes

            # If a required scope is provided, we first check to see whether
            # the session globally or explicitly authorizes the request. We
            # then fall back to the locally-defined authorizer function if it
            # is provided.
            if required and scopes:

                # A global scope is usually granted to administrators, or
                # perhaps moderators (e.g. view submission content).
                # For example: `submission:read:*`.
                if required.as_global() in scopes:
                    logger.debug('Authorized with global scope')
                    authorized = True

                # A resource-specific scope may be granted at the auth layer.
                # For example, an admin may provide provisional access to a
                # specific resource for a specific role. This kind of
                # authorization is only supported if the service provides a
                # ``resource()`` callback to get the resource identifier.
                elif (resource is not None
                      and (
                        required.for_resource(str(resource(*args, **kwargs)))
                        in scopes)):
                    logger.debug('Authorized by specific resource')
                    authorized = True

                # If both the global and resource-specific scope authorization
                # fail, then we look for the general scope in the session.
                elif required in scopes:
                    # If an authorizer callback is provided by the service,
                    # then we will enforce whatever it returns.
                    if authorizer:
                        authorized = authorizer(session, *args, **kwargs)
                    # If no authorizer callback is provided, it is implied that
                    # the general scope is sufficient to authorize the request.
                    elif authorizer is None:
                        authorized = True
                # The required scope is not present. There is nothing left to
                # check.
                else:
                    authorized = False

            elif required is None and authorizer is None:
                logger.debug('No scope required, no authorizer function;'
                             ' request is authorized.')
                authorized = True

            # If a specific scope is not required, we rely entirely on the
            # authorizer callback.
            elif authorizer is not None:
                logger.debug('Calling authorizer callback')
                authorized = authorizer(session, *args, **kwargs)

            if not authorized:
                logger.debug('Session is not authorized for %s', required)
                raise Forbidden('Access denied')  # type: ignore

            logger.debug('Request is authorized, proceeding')
            return func(*args, **kwargs)
        return wrapper
    return protector
